#include <windows.h>
#include <winnt.h>

#include "debugger.h"
#include "Psapi.h"

#define IKWYD
#include "config.h"
#include "common/debug.h"
#include "common/memory.h"

#include "driver.h"


static char g_param1[260];
HANDLE g_hFile = NULL;
static CHAR szDebuggeeDir[MAX_PATH+1];

/*
Gets double word from process process. Returns true on successful.
*/
static BOOL ReadDWORDFromProcess ( HANDLE process, PVOID addr, PDWORD res )
{
	DWORD total = 0, rc;
	*res = 0;
	rc = ReadProcessMemory ( process, addr, res, 4, &total );
	if ( total != 4 ) rc = 0;
	return rc;
}

#define BUFSIZE 512

/**
 * @bref Object name to file name
 * @param in pszObjectName - object name
 * @param out pszFileName - file name
 *
 * @return process status
 *	@value 0 - sucessfully complete
 */
static int TranslateToFilePath ( PCHAR pszObjectName, PCHAR pszFileName )
{
	TCHAR szTemp[BUFSIZE];
	szTemp[0] = '\0';
	BOOL bFound = FALSE;

	if ( GetLogicalDriveStrings ( BUFSIZE - 1, szTemp ) )
	{
		CHAR szName[ MAX_PATH ];
		CHAR szDrive[3] = TEXT ( " :" );
		CHAR *p = szTemp;

		do
		{
			// Copy the drive letter to the template string
			*szDrive = *p;

			// Look up each device name
			if ( QueryDosDevice ( szDrive, szName, BUFSIZE ) )
			{
				UINT uNameLen = strlen ( szName );

				if ( uNameLen < MAX_PATH )
				{
					bFound = _strnicmp ( pszObjectName, szName,
										 uNameLen ) == 0;

					if ( bFound )
					{
						// Reconstruct pszFilename using szTempFile
						// Replace device path with DOS path
						TCHAR szTempFile[MAX_PATH];
						StringCchPrintf ( szTempFile,
										  MAX_PATH,
										  TEXT ( "%s%s" ),
										  szDrive,
										  pszObjectName + uNameLen );
						StringCchCopyN ( pszFileName, MAX_PATH + 1, szTempFile, strlen ( szTempFile ) + 1 );
					}
				}
			}

			// Go to the next NULL character.
			while ( *p++ );
		}
		while ( !bFound && *p ); // end of string
	}

	return !bFound;
}

/**
 * @bref Get loaded module full name
 *
 * @param in pIbase, hFile, lpImageName, fUnicode - debug object parameters
 * @param out pszModuleName - module name
 *
 * @return - procedure status
 *	@value 0 - successfully complete
 *	@value -1 - can't get module name
 */
static int GetModuleName ( HINSTANCE pIbase, HANDLE hFile, PVOID lpImageName, BOOL fUnicode, PCHAR pszModuleName )
{
	PCHAR pModuleName;

	// Module name can be ANSI or UNOCODE
	wchar_t szUnicodeModuleName[MAX_PATH+1];
	CHAR szCharModuleMame[MAX_PATH+1];
	SIZE_T NumberOfBytesRead;

	// Remote thread handler
	POBJECT_NAME_INFORMATION	pfni;
	DWORD nbread;
	CHAR szName[MAX_PATH];

	// Completeness flag
	BOOL IsName = false;
	BOOL IsUnicodeName = false;

	// Read module name from de structure
	if ( lpImageName
			&& ReadDWORDFromProcess ( hFile, lpImageName, ( PDWORD ) &pModuleName )
			&& pModuleName != NULL )
	{
		if ( fUnicode )
		{
			ReadProcessMemory ( hFile, pModuleName, szUnicodeModuleName, MAX_PATH*2, &NumberOfBytesRead );
			if ( NumberOfBytesRead != 0 )
			{
				IsUnicodeName = true;
				IsName = true;
			}
		}
		else
		{
			ReadProcessMemory ( hFile, pModuleName, szCharModuleMame, MAX_PATH, &NumberOfBytesRead );
			if ( NumberOfBytesRead != 0 )
			{
				IsUnicodeName = false;
				IsName = true;
			}
		}
	}

	// try another way
	if ( !IsName && GetModuleBaseName ( hFile, ( HINSTANCE ) pIbase, szCharModuleMame, sizeof ( szName ) ) )
	{
		IsUnicodeName = false;
		IsName = true;
	}

	if ( !IsName )
	{
		pfni = ( POBJECT_NAME_INFORMATION ) GlobalAlloc ( GPTR, MAX_FNI_LENGTH * 2 );
		pfni->Name.Buffer = ( PWSTR ) ( ( DWORD ) pfni + sizeof ( OBJECT_NAME_INFORMATION ) );
		memset ( szName, 0, MAX_PATH );

		if ( ( LONG ) ( NtQueryObject ( ( HANDLE ) hFile, ObjectNameInformation, pfni, MAX_FNI_LENGTH, &nbread ) >= 0 ) )
		{
			CHAR pszTmpName[ MAX_PATH ];

			WideCharToMultiByte ( CP_ACP, 0, pfni->Name.Buffer, pfni->Name.Length >> 1,
								  pszTmpName, MAX_PATH, NULL, NULL );

			pfni->Name.Length = pfni->Name.Length >> 1;
			pszTmpName[ pfni->Name.Length ] = '\0';

			if ( !TranslateToFilePath ( pszTmpName, szCharModuleMame ) )
			{
				IsUnicodeName = false;
				IsName = true;
			}
		}

		GlobalFree ( pfni );
	}

	if ( !IsName )
	{
		printf ( "can't get module name\n\n" );
		return -1;
	}

	if ( IsUnicodeName )
	{
		WideCharToMultiByte ( CP_ACP, 0, szUnicodeModuleName, NumberOfBytesRead >> 1,
							  pszModuleName, MAX_PATH, NULL, NULL );
	}
	else
		lstrcpyn ( pszModuleName, ( PCHAR ) szCharModuleMame, MAX_PATH );

	return 0;
}


#define DebugCtlMSR 0x01D9
void SetTfFlag ( DWORD dwThreadID )
{
	HANDLE hThread;
	CONTEXT cn;
	hThread = OpenThread ( THREAD_GET_CONTEXT | THREAD_SET_CONTEXT, false, dwThreadID );

	if ( ReadContext ( hThread, &cn ) )
	{
		cn.EFlags |= 0x0100; //bit Tf
		SetThreadContext ( hThread, &cn );
	}

	CloseHandle ( hThread );
}

static int DE_DebugError ( DEBUG_EVENT *de )
{
	LogDbg( "DE_DebugError() [%X]: error=%i, type=%i\n",de->dwThreadId,de->u.RipInfo.dwError,de->u.RipInfo.dwType);
	return DBG_CONTINUE;
}


void ResetTfFlag ( DWORD dwThreadID )
{
	HANDLE hThread;
	CONTEXT cn;
	hThread = OpenThread ( THREAD_GET_CONTEXT | THREAD_SET_CONTEXT, false, dwThreadID );

	if ( ReadContext ( hThread, &cn ) )
	{
		cn.EFlags &= 0xFEFF; //!bit Tf
		SetThreadContext ( hThread, &cn );
	}

	CloseHandle ( hThread );
}

static int ProcessSingleStep ( DEBUG_EVENT *de )
{
	LogInfo( "!0x%x\n", de->u.Exception.ExceptionRecord.ExceptionAddress );
	return DBG_CONTINUE;
}


// with break point, had consist with two instruction (push retadr, jmp funcadr)
static int DE_Exception ( DEBUG_EVENT *de )
{
	int retVal = DBG_CONTINUE;

	switch ( de->u.Exception.ExceptionRecord.ExceptionCode )
	{
		case ( EXCEPTION_SINGLE_STEP ) :
			retVal = ProcessSingleStep ( de );
			break;
		default:
			LogDbg ( "Exception unknown at address %x with exception code %x\n",
					 ( ( DWORD ) de->u.Exception.ExceptionRecord.ExceptionAddress ), de->u.Exception.ExceptionRecord.ExceptionCode );
			retVal = DBG_CONTINUE; // nevermind
	}

	return retVal;
}

static int DllLoadUnload( DEBUG_EVENT *de, bool isLoaded )
{
	char	pszModuleName[MAX_PATH];
	
	if ( GetModuleName ( ( HINSTANCE ) de->u.CreateProcessInfo.lpBaseOfImage,
		de->u.CreateProcessInfo.hFile,
		de->u.CreateProcessInfo.lpImageName, de->u.CreateProcessInfo.fUnicode, pszModuleName ) )
	{
		LogErr ( "Can't get module information\n" );
		return -1;
	}
	
	_strlwr ( pszModuleName );
	
	LogInfo ( "Module %s is %s, base 0x%x\n" , pszModuleName, isLoaded?"loaded":"unloaded", isLoaded? de->u.LoadDll.lpBaseOfDll:de->u.UnloadDll.lpBaseOfDll );
	
	return 0;
}

static int DE_LoadDll ( DEBUG_EVENT *de )
{
	if( DllLoadUnload( de, true ) )
	{
		LogWrn( "Can't get module name\n" );
		return DBG_CONTINUE;
	}

	return DBG_CONTINUE;
}

static int DE_UnloadDll ( DEBUG_EVENT *de )
{
	if( DllLoadUnload( de, false ) )
	{
		LogWrn( "Can't get module name\n" );
		return DBG_CONTINUE;
	}

	return DBG_CONTINUE;
}

int StartProcess( char * command, STARTUPINFO * si, PROCESS_INFORMATION	* pi )
{
	memset ( si, 0, sizeof ( STARTUPINFO ) );\
	si->cb = sizeof ( STARTUPINFO );

	if ( !CreateProcess ( NULL, command, NULL, NULL, FALSE,
						  DEBUG_PROCESS | CREATE_NEW_CONSOLE,
						  NULL, NULL, si, pi ) )
	{
		LogVrb ( "CreateProcess() failed with LastError=%i\n", GetLastError() );
		return -1;
	}

	return 0;
}

void main ( int argc, PCHAR argv[] )
{
	DWORD	dwContinueStatus, i, dwArgsSize;
	PCHAR	pszArgs;
	DEBUG_EVENT	de;
	
	STARTUPINFO	si;
	PROCESS_INFORMATION	pi;

	LogInfo ( "IKWYD Debugger v"CF_VERSION" \n\n" );
	if ( ( argc < 2 ) || !strcmp( argv[1], "/?" ) )
	{
		printf ( "Usage: %s [ program.exe [arguments] | --uninstall | /? ]\n", argv[0] );
		return;
	}

	if ( ( argc >= 1 ) && !strcmp( argv[1], "--uninstall" ) )
	{
		LogInfo( "Need to uninstall driver\n" );
		
		if( UnloadDriver( "setbdf", &g_hFile ) )
		{
			LogErr( "Can't uninstall driver setbdf.sys\n" );
		}else
		{
			LogInfo( "Driver setbdf.sys uninstalled\n" );
		}

		return;
	}

	// Initialization
	if( LoadDriver( "setbdf", &g_hFile ) )
	{
		LogErr( "Can't load driver setbdf.sys\n" );
		return;
	}

	// collect debuggee arguments
	for ( i = 1, dwArgsSize = 0;i < ( DWORD ) argc;i++ ) dwArgsSize += lstrlen ( argv[i] );
	pszArgs = ( PCHAR ) GlobalAlloc ( GPTR, dwArgsSize << 2 );
	if ( !pszArgs )
	{
		LogErr ( "Can't allocate memory for arguments\n" );
		return;
	}

	for ( i = 1;i < ( DWORD ) argc;i++ )
	{
		lstrcat ( pszArgs, argv[i] );
		if ( i != ( DWORD ) argc - 1 ) lstrcat ( pszArgs, " " );
	}
	
	// Start process
	if( !StartProcess( pszArgs, &si, &pi ) )
	{
		LogDbg( "Pocess successfuly executed\n" );
		LogDbg( "State changed to EXECUTING\n" );
	} else
	{
		LogErr( "Can't start %s\n", pszArgs );
		return;
	}

	bool bMustQuit = false;
	bool needSetTF = false;

	while ( !bMustQuit )
	{	
			if ( WaitForDebugEvent ( &de, 50 ) != NULL )
			{
				switch ( de.dwDebugEventCode )
				{
					case CREATE_PROCESS_DEBUG_EVENT:
						dwContinueStatus = DBG_CONTINUE;
						LogDbg ( "Process with ID 0x%x started.\n", de.dwProcessId );
						needSetTF = true;
						break;
					case CREATE_THREAD_DEBUG_EVENT:
						dwContinueStatus = DBG_CONTINUE;
						LogDbg ( "Thread with ID 0x%x started!\n", de.dwThreadId );
						needSetTF = true;
						break;
					case EXIT_PROCESS_DEBUG_EVENT:
						dwContinueStatus = DBG_CONTINUE;
						LogDbg ( "Process with ID 0x%x terminated!\n", de.dwThreadId );
						if( de.dwProcessId == pi.dwProcessId )
						{
							LogDbg( "Need to TERMINATE\n" );
							bMustQuit = true;
						}
						break;
					case EXIT_THREAD_DEBUG_EVENT:
						dwContinueStatus = DBG_CONTINUE;
						LogDbg ( "Thread with ID 0x%x exited!\n", de.dwThreadId );
						break;	
					case RIP_EVENT:
						dwContinueStatus = DE_DebugError ( &de );
						break;
					case EXCEPTION_DEBUG_EVENT:
						dwContinueStatus = DE_Exception ( &de );
						if( de.u.Exception.ExceptionRecord.ExceptionCode == EXCEPTION_SINGLE_STEP ||
							de.u.Exception.ExceptionRecord.ExceptionCode == EXCEPTION_BREAKPOINT )
							needSetTF = true;
						break;
					case LOAD_DLL_DEBUG_EVENT:
						dwContinueStatus = DE_LoadDll ( &de );
						dwContinueStatus = DBG_CONTINUE;
						break;
					case UNLOAD_DLL_DEBUG_EVENT:
						dwContinueStatus = DE_UnloadDll ( &de );
						dwContinueStatus = DBG_CONTINUE;
						break;
					default:
						LogVrb ( "Weird dwDebugEventCode (%i)\n", de.dwDebugEventCode );
						dwContinueStatus = DBG_CONTINUE;
				}

				if ( needSetTF )
					SetTfFlag ( de.dwThreadId );

				ContinueDebugEvent ( de.dwProcessId, de.dwThreadId, dwContinueStatus );
			}
		}
}

