/**
 *  Определено пять уровней сообщений: MSG_FAULT, MSG_ERROR, MSG_WARNING, MSG_DEBUG, MSG_VERBOUSE.
 *  Каждый уровень включает логирование предыдущего. Логирование FAULT осуществляеся всегда, если
 *  определен макрос DEBUG. Данный файл включатся совместно с файлом config.h, в котором
 *  указывается имя логируемого модуля и уровень логирования.
 *  Пример использования в произвольном модуле (код в модуле )
 *  ---------aaa.c-------------
 *  #define MODULE_AAAA
 *  #include "config.h"
 *  #include "debug.h"
 *  ------секция в config.h----
 *  #ifdef MODULE_AAAA
 *  #define MOD_NAME "module_aaa"
 *  #define MSG_WARNING
 *  #endif
 *  ---------------------------
 */

#include <stdio.h>
#ifndef __DEBUG_H__
#define __DEBUG_H__

// this macros set in project
#define _DEBUG
#ifdef _DEBUG
#define _DRIVERNAME_    "Dataflow"

#ifdef MSG_VERBOUSE
#define MSG_DEBUG
#define MSG_INFO
#define MSG_WARNING
#define MSG_ERROR 
#define LogVrb( format, ... ) \
	printf( "%s:%s, (-,%d) %s(): " format, _DRIVERNAME_, MOD_NAME,  __LINE__, __FUNCTION__, __VA_ARGS__ ); 
#else
#define	LogVrb(format, ... ) \
	do {} while(0)
#endif 

#ifdef MSG_DEBUG
#define MSG_WARNING 
#define MSG_ERROR 
#define MSG_INFO
#define LogDbg(format, ...) \
	printf( "%s:%s, (-,%d) %s(): " format, _DRIVERNAME_, MOD_NAME,  __LINE__, __FUNCTION__, __VA_ARGS__ );
#else
#define LogDbg(format, ...) \
	do {} while(0)
#endif 

#ifdef MSG_WARNING
#define MSG_ERROR
#define MSG_INFO
#define LogWrn( format, ... ) \
	printf( "%s:%s, (-,%d) %s(): Warning! " format, _DRIVERNAME_, MOD_NAME,  __LINE__, __FUNCTION__, __VA_ARGS__ );
#else
#define LogWrn( format, ... ) \
	do {} while(0)
#endif 

#ifdef MSG_ERROR
#define MSG_INFO
#define LogErr( format, ... ) \
	printf( "%s:%s, (-,%d) %s(): Error! " format, _DRIVERNAME_, MOD_NAME,  __LINE__, __FUNCTION__, __VA_ARGS__ );
#else
#define LogErr( format, ... ) \
	do {} while(0)
#endif 

#ifdef MSG_INFO 
#define LogInfo( format, ... ) \
	printf( format, __VA_ARGS__ );
#else
#define LogInfo( format, ... ) \
	do {} while(0)
#endif 

#define LogFlt( format, ... ) \
	printf( "%s:%s, (-,%d) %s(): Fault! " format, _DRIVERNAME_, MOD_NAME, __LINE__, __FUNCTION__, __VA_ARGS__ );
	
#define ASSERT(expr) \
	if(!(expr)) \
	{ \
		LogFlt( "Assertion failed: %s\n", #expr ); \
		__asm{ \
		int 3 \
		}	\
	}	
	
#else // DBG
	
#define	ASSERT(expr) \
	do {} while(0)

#define LogVrb( format, ... ) \
	do {} while(0)
#define LogWrn( format, ... ) \
	do {} while(0)
#define LogDbg(format, ...) \
	do {} while(0)
#define LogInfo(format, ...) \
	do {} while(0)
#define LogErr( format, ... ) \
	do {} while(0)
#define LogFlt(format, ...) \
	do {} while(0)

#define ASSERT(expr) \
	do {} while(0)

#endif // DBG

#endif //_DEBUG_H
