#define MEM_WRAPPER
#include "config.h"

#include "memory.h"
#include <stdlib.h>

#include "common/debug.h"

#ifdef MEM_DEBUG

struct area_descriptor 
{
	struct area_descriptor * next;
	void * begin;
	size_t size;
	unsigned long line;
	char fname[255];
};

static struct area_descriptor * g_head = NULL;

unsigned long gMaxMemory = 0;
unsigned long gCurrentMemory = 0;

HRESULT __AllocateMemory( void ** ptr, ssize_t size,  unsigned long line, char * fileName)
{
	struct area_descriptor * area;

	area = malloc ( sizeof ( struct area_descriptor ) );
	if ( !area )
	{
		printf ( "Allocation error\n" );
		return HR_ALLOCATION_ERROR;
	}

	*ptr = malloc ( size );
	if ( !*ptr )
	{
		printf ( "Allocation error\n" );
		return HR_ALLOCATION_ERROR;
	}
	
	area->begin = *ptr;
	area->size = size;
	area->line = line;
	area->next = g_head;

	strncpy ( area->fname, file, 255 );

	g_head = area;

	//printf ( "%s: end\n", __FUNCTION__ );
	gCurrentMemory += size;
	if( gCurrentMemory > gMaxMemory )
		gMaxMemory = gCurrentMemory;

	return HR_SUCCESS;
}

HRESULT __FreeMemory ( void ** ptr, unsigned long line, char * file )
{
	struct area_descriptor * area = g_head, * prev = NULL;

	//printf ( "%s: start\n", __FUNCTION__ );
	while ( area )
	{
		if ( area->begin == *ptr )
		{
			gCurrentMemory -= area->size;

			if ( prev )
				prev->next = area->next;
			else
				g_head = area->next;
			free ( area->begin );
			free ( area );
			//printf ( "%s: end\n", __FUNCTION__ );
			*ptr = NULL;
		

			return HR_SUCCESS;
		}
		prev = area;
		area = area->next;
	}

	printf ( "In file %s at line %ld try to free not allocated memory\n", file, line );
	//free ( *ptr );

	*ptr = NULL;

	return HR_FAILED;
}

HRESULT __FreeAllAllocatedMemory ( void )
{
	struct area_descriptor * tmp;
	int retval = HR_SUCCESS;
	
	if ( !g_head )
		retval = HR_SUCCESS;

	while ( g_head )
	{
		tmp = g_head;
		printf ( "Memory chunk:\n\tbegin from %p size %d\n\tallocated at line %ld in file:\n\t%s\n", 
				g_head->begin, ( int ) g_head->size, g_head->line, g_head->fname ); 
		free ( g_head->begin );
		g_head = g_head->next;
		free ( tmp );
	}

	printf( "Max allocated memory amount %lu bytes\n", gMaxMemory );

	return retval;
}

#else
/**
 * @fn HRESULT __AllocateMemory( void ** ptr, ssize_t size, gfp_t gfp, unsigned long line, char * fileName)
 * @param ptr[out] Указатель в котором возвращаем начало выделенной области
 * @param size[in] Размер выделяемой области
 * @param line[in] Номер строки где произведен вызов
 * @param fileName[in] Имя файла где произведен вызов
 * @return HR_SUCCESS если все прошло нормально, HR_ALLOCATION_ERROR невозможно выделить участок памяти 
 * @return данного размера, HR_BAD_ARGUMENTS если переданны неверные аргументы.
 * @brief Выделение памяти с логированием, если собирали с MEM_DEBUG. Иначе просто выделение памяти.
 */
/*  */ HRESULT __AllocateMemory( void ** ptr, size_t size )
{
	*ptr = malloc( size);
	if ( ptr == NULL )
	{
		return HR_ALLOCATION_ERROR;
	}
	return HR_SUCCESS;
}

/**
 * @fn HRESULT __FreeMemory( void ** ptr , unsigned long line, char * fileName )
 * @param ptr Указатель на начало участка памяти.
 * @param line Номер строки в котором вызывается
 * @param fileName имя файла в котором вызывается
 * @return Если выделяли этот участок памяти HR_SUCCESS, иначе HR_FAILED.
 * @brief Освобождение памяти выделенной с помощью AllocateMemory
 */
/*  */ HRESULT __FreeMemory( void ** ptr )
{
	free( *ptr );
	*ptr = NULL;
	return HR_SUCCESS;
}

#endif // MEM_DEBUG
