#ifndef __memory_h
#define __memory_h

#define MEM_WRAPPER
#include "config.h"

// Функции для работы с памятью наподобие memset и memcpy будут отличаться
// для разных платформ/колец. Здесь подключим stdlib. В других платформах
// прийдется как то решать проблему.
#include <string.h>
#include <strsafe.h>

#define HR_FALSE                       -1
#define HR_SUCCESS						0
#define HR_BAD_ARGUMENTS 				1
#define HR_ALLOCATION_ERROR             2
#define HR_FAILED						3
#define HR_MEMORY_UNCLEAN				4

typedef int ssize_t;

#ifdef __cplusplus
extern "C"
{
#endif

#ifndef MEM_DEBUG
//Функции без MEM_DEBUG
/* inline */ HRESULT __AllocateMemory( void ** ptr, size_t size );
/* inline */ HRESULT __FreeMemory( void ** ptr );

#define AllocateMemory( ptr, size ) __AllocateMemory( ( void ** ) &ptr, size )
#define AllocateGFPMemory AllocateMemory
#define AllocateVirtualMemory AllocateMemory
#define FreeMemory( ptr ) __FreeMemory( ( void ** ) &ptr )
#define FreeVirtualMemory FreeMemory
#define FreeAllAllocatedMemory( )

#else

HRESULT __AllocateMemory( void ** ptr, ssize_t size,  unsigned long line, char * fileName);
HRESULT __FreeMemory( void ** ptr, unsigned long line, char * fileName );
HRESULT __FreeAllAllocatedMemory( void );

#define AllocateMemory( ptr, size ) __AllocateMemory( (void * ) &ptr, size, __LINE__, __FILE__ )
#define AllocateGFPMemory( ptr, size ) __AllocateMemory( (void * ) &ptr, size, __LINE__, __FILE__ )
#define AllocateVirtualMemory( ptr, size ) __AllocateMemory( (void * ) &ptr, size, __LINE__, __FILE__ )
#define FreeMemory( ptr ) __FreeMemory( ( void * ) &ptr, __LINE__, __FILE__ )
#define FreeVirtualMemory( ptr ) __FreeMemory( ( void * ) &ptr, __LINE__, __FILE__ )
#define FreeAllAllocatedMemory( ) __FreeAllAllocatedMemory()

#endif // MEM_DEBUG

#ifdef __cplusplus
}
#endif

#endif /* __memwraps_h */
