#include <ntddk.h>
#include <stdio.h>
#include <windef.h>

#define NT_DEVICE_NAME      L"\\Device\\setbdf"
#define DOS_DEVICE_NAME     L"\\DosDevices\\setbdf"


__drv_dispatchType(IRP_MJ_DEVICE_CONTROL)
DRIVER_DISPATCH SetBdf_IoControl;

PDEVICE_OBJECT  g_deviceObject = NULL;

/* sidt returns idt in this format */
typedef struct

{
   unsigned short IDTLimit;

   unsigned short LowIDTbase;

   unsigned short HiIDTbase;

} IDTINFO;


// entry in the IDT: this is sometimes called
// an "interrupt gate"
#pragma pack(1)

typedef struct

{

      unsigned short LowOffset;

      unsigned short selector;

      unsigned char unused_lo;

      unsigned char segment_type:4; //0x0E is interrupt gate

      unsigned char system_segment_flag:1;

      unsigned char DPL:2;     // descriptor privilege level

      unsigned char P:1;       // present

      unsigned short HiOffset;

} IDTENTRY;

#pragma pack()

//#define MAKELONG(a, b) ((unsigned long) (((unsigned short) (a)) | ((unsigned long) ((unsigned short) (b))) << 16))

DWORD  KiDebugRoutineISR_Ptr; // The real INT 2E handler
//DWORD  ReturnAddress;
//DWORD  CSRegister;
//DWORD  FlagsRegister;

#define NT_DEBUG_INT 0x01

static DWORD g_lastBranchEIP = 0;
static DWORD g_lastBranchEIPHigh = 0;

__declspec(naked) MyKiDebugRoutine()

{
	__asm{
		pushad
		pushfd
		push fs
		mov bx, 0x30
		mov fs, bx
   }
		
   //DbgPrint( "I'm in int1 now!\n" );
  
   __asm{
	    mov eax, dr6
		bt eax, 14
		jnc not_single_step

		//save last branch
		mov ecx, 01C9h // LastBranchFromIP
		rdmsr
		and eax, 0fh 
		dec eax		// module f add
		and eax, 0fh
		
		mov ecx, 680h
		add ecx, eax
		rdmsr

		mov g_lastBranchEIP, eax
		mov g_lastBranchEIPHigh, edx
   }
		//DbgPrint( "last branch %x-%x\n", g_lastBranchEIPHigh, g_lastBranchEIP );
	
		__asm{
	    mov ecx, 1D9h // MSR_DEBUGCTLA
		rdmsr //MOV (EDX:EAX), MSR_DEBUGCTLA
		or eax, 3 // SET BTF (single-step on branches) flag (bit 1)
		wrmsr // MOV (EDX:EAX), MSR_DEBUGCTLA

not_single_step:
   }
   
   __asm{	    
		//int 3
	    pop fs
		popfd
		popad
	
		//mov eax, [esp + 4]
		//mov [ReturnAddress], eax
		//mov eax, [esp + 8]
		//mov [CSRegister], eax
		//mov eax, [esp + 0xc]
		//mov [FlagsRegister], eax
		//and eax, 0xFFFFFEFF
		//mov [esp + 0xc ], eax
		//pop eax
		//add esp, 8
		//push cs
		//call   [KiDebugRoutineISR_Ptr];  // Call the real function
		//cli
		//push [CSRegister]
		//push [FlagsRegister]
		//push [ReturnAddress]
		//sti
		//iretd
		
		//jmp KiDebugRoutineISR_Ptr

		jmp KiDebugRoutineISR_Ptr
   }
}

int UnhookInterrupts()
{
   IDTINFO idt_info;
   IDTENTRY* idt_entries;
   IDTENTRY* int1_entry;

   __asm{
      sidt idt_info;
   }

   idt_entries = ( IDTENTRY * )MAKELONG( idt_info.LowIDTbase, idt_info.HiIDTbase );

   /*******************************************************
    * Note: we can patch ANY interrupt here;
    * the sky is the limit
    *******************************************************/

   int1_entry = &(idt_entries[NT_DEBUG_INT]);

   __asm{

     cli;                       // Mask Interrupts
     mov eax, [KiDebugRoutineISR_Ptr];  // Load EAX with the address of
                                // hook
     mov ebx, int1_entry;      // Address of INT 2E handler in
                                // table
     mov [ebx],ax;              // Overwrite real handler with
                                // the low
                                // 16 bits of the hook address.
     shr eax,16

     mov [ebx+6],ax;           // Overwrite real handler with
                               // the high
                               // 16 bits of the hook address.
     sti;                      // Enable Interrupts again.
   }

}

int HookInterrupts()
{

   IDTINFO idt_info;
   IDTENTRY* idt_entries;
   IDTENTRY* int1_entry;

   __asm{
      sidt idt_info;
   }

   idt_entries = ( IDTENTRY * )MAKELONG( idt_info.LowIDTbase, idt_info.HiIDTbase );

   // Save the real address of the handler.
   KiDebugRoutineISR_Ptr = MAKELONG( idt_entries[NT_DEBUG_INT].LowOffset,
         idt_entries[NT_DEBUG_INT].HiOffset );

   /*******************************************************
    * Note: we can patch ANY interrupt here;
    * the sky is the limit
    *******************************************************/

   int1_entry = &(idt_entries[NT_DEBUG_INT]);

   __asm{

     cli;                       // Mask Interrupts
     lea eax,MyKiDebugRoutine;  // Load EAX with the address of
                                // hook
     mov ebx, int1_entry;      // Address of INT 2E handler in
                                // table
     mov [ebx],ax;              // Overwrite real handler with
                                // the low
                                // 16 bits of the hook address.
     shr eax,16

     mov [ebx+6],ax;           // Overwrite real handler with
                               // the high
                               // 16 bits of the hook address.
     sti;                      // Enable Interrupts again.
   }

   return 0;
}

#define IOCTL_SETBDF 60

#define IOCTL_GET_LAST_BRANCH_IO     \
        CTL_CODE(IOCTL_SETBDF,        \
                 0x0, \
				 METHOD_BUFFERED, \
                 FILE_ANY_ACCESS )

NTSTATUS GetLastBranchIOCTL( PIRP Irp, 
        PIO_STACK_LOCATION pIoStackIrp, UINT *pdwDataWritten )
{
    NTSTATUS NtStatus = STATUS_UNSUCCESSFUL;
    PCHAR pInputBuffer;
    PCHAR pOutputBuffer;
    UINT dwDataRead = 0, dwDataWritten = 0;
    UINT dwDataSize = sizeof( DWORD );

    //DbgPrint("GetLastBranchIOCTL Called \r\n");

    /*
     * METHOD_BUFFERED
     *
     *    Input Buffer = Irp->AssociatedIrp.SystemBuffer
     *    Ouput Buffer = Irp->AssociatedIrp.SystemBuffer
     *
     *    Input Size   =  Parameters.DeviceIoControl.InputBufferLength
     *    Output Size  =  Parameters.DeviceIoControl.OutputBufferLength
     *
     *    Since they both use the same location
     *    so the "buffer" allocated by the I/O
     *    manager is the size of the larger value (Output vs. Input)
     */


    pInputBuffer = ( PCHAR ) Irp->AssociatedIrp.SystemBuffer;
    pOutputBuffer = ( PCHAR ) Irp->AssociatedIrp.SystemBuffer;

    if( pInputBuffer && pOutputBuffer )
    {                                                      
        //DbgPrint("Current EIO: %x\n", *( (DWORD *) Irp->AssociatedIrp.SystemBuffer) );
        //DbgPrint("%i >= %i, lastBranchEIP: %x\n", 
        //   pIoStackIrp->Parameters.DeviceIoControl.OutputBufferLength, 
        //    dwDataSize, g_lastBranchEIP);

        if(pIoStackIrp->Parameters.DeviceIoControl.OutputBufferLength
                                                        >= dwDataSize)
        {
            * (( DWORD * ) pOutputBuffer) = g_lastBranchEIP;
            *pdwDataWritten = dwDataSize;
            NtStatus = STATUS_SUCCESS;
        }
        else
        {
            *pdwDataWritten = dwDataSize;
            NtStatus = STATUS_BUFFER_TOO_SMALL;
        }
    }

    return NtStatus;
}

NTSTATUS SetBdf_IoControl(PDEVICE_OBJECT DeviceObject, PIRP Irp)
{
    NTSTATUS NtStatus = STATUS_NOT_SUPPORTED;
    PIO_STACK_LOCATION pIoStackIrp = NULL;
    UINT dwDataWritten = 0;

    //DbgPrint( "IOCTL_GET_LAST_BRANCH_IO Called \r\n" );

    pIoStackIrp = IoGetCurrentIrpStackLocation( Irp );    

    if( pIoStackIrp ) /* Should Never Be NULL! */
    {
        switch( pIoStackIrp->Parameters.DeviceIoControl.IoControlCode )
        {
            case IOCTL_GET_LAST_BRANCH_IO :
                 NtStatus = GetLastBranchIOCTL(Irp, 
                                 pIoStackIrp, &dwDataWritten);
                break;

			default:
				//DbgPrint( "Unknown IOCTL %i\n", pIoStackIrp->Parameters.DeviceIoControl.IoControlCode );
				break;
        }
    }

    Irp->IoStatus.Status = NtStatus;
    Irp->IoStatus.Information = dwDataWritten;

    IoCompleteRequest(Irp, IO_NO_INCREMENT);

    return NtStatus;
}

NTSTATUS
SioctlCreateClose(
    __in PDEVICE_OBJECT DeviceObject,
    __in PIRP Irp
    )
/*++

Routine Description:

    This routine is called by the I/O system when the SIOCTL is opened or
    closed.

    No action is performed other than completing the request successfully.

Arguments:

    DeviceObject - a pointer to the object that represents the device
    that I/O is to be done on.

    Irp - a pointer to the I/O Request Packet for this request.

Return Value:

    NT status code

--*/

{
    PAGED_CODE();

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}


VOID OnUnload( IN PDRIVER_OBJECT DriverObject )
{
	UNICODE_STRING ntWin32NameString = RTL_CONSTANT_STRING(DOS_DEVICE_NAME);
	UnhookInterrupts();

	 IoDeleteSymbolicLink( &ntWin32NameString );

	if ( g_deviceObject != NULL )
	{
		IoDeleteDevice( g_deviceObject );
	}

	DbgPrint("Driver unloaded\n");

}

NTSTATUS DriverEntry( IN PDRIVER_OBJECT pDriverObject,
                     IN PUNICODE_STRING theRegistryPath )
{
	NTSTATUS status = STATUS_SUCCESS;
	UNICODE_STRING ntUnicodeString = RTL_CONSTANT_STRING(NT_DEVICE_NAME);
	UNICODE_STRING ntWin32NameString = RTL_CONSTANT_STRING(DOS_DEVICE_NAME);

	HookInterrupts();

	status = IoCreateDevice(
		pDriverObject,                  // Our Driver Object
		0,                              // We don't use a device extension
		&ntUnicodeString,               // Device name "\Device\SIOCTL"
		FILE_DEVICE_UNKNOWN,					// Device type
		FILE_DEVICE_SECURE_OPEN,        // Device characteristics
		FALSE,                          // Not an exclusive device
		&g_deviceObject );                // Returned ptr to Device Object
						
	if ( !NT_SUCCESS( status ) )
	{
		//DbgPrint("Couldn't create the device object\n");
		return status;
	}

	status = IoCreateSymbolicLink(&ntWin32NameString, &ntUnicodeString );
	if(!NT_SUCCESS(status))
		DbgPrint("Can't Create Symbol link %x\n", status);
	
	pDriverObject->DriverUnload = OnUnload;
	pDriverObject->MajorFunction[IRP_MJ_CREATE] = SioctlCreateClose;
    pDriverObject->MajorFunction[IRP_MJ_CLOSE] = SioctlCreateClose;
	pDriverObject->MajorFunction[ IRP_MJ_DEVICE_CONTROL ] = SetBdf_IoControl;

	DbgPrint("Driver loaded\n");

	return STATUS_SUCCESS;
}

